#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>

#include "lbx.h"
#include "image.h"

static GtkTreeStore *archives;
static GtkBuilder *builder;

static GtkWidget *canvas;
static LBX_IMG *image;
/* LBX images can have up to three palettes, with each superseding the last. */
static struct lbx_colour palette_external[256];
static struct lbx_colour palette_internal[256];
static struct lbx_colour palette_override[256];

GdkPixbuf *framebuf;

void play_toggled(GtkToggleButton *button, gpointer data)
{
	gboolean active = gtk_toggle_button_get_active(button);
	GtkWidget *spin;

	if (active && !image) {
		gtk_toggle_button_set_active(button, FALSE);
		return;
	}

	spin = GTK_WIDGET(gtk_builder_get_object(builder, "framespin"));
	gtk_widget_set_sensitive(spin, !active);

	if (active) {
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin), 0);
	}
}

static void get_colour(unsigned char index, unsigned char out[static 3])
{
	struct lbx_colour *colour;

	if (palette_override[index].active)
		colour = palette_override + index;
	else if (palette_internal[index].active)
		colour = palette_internal + index;
	else if (palette_external[index].active)
		colour = palette_external + index;
	else
		colour = &(struct lbx_colour) { .red = 0xff, .blue = 0xff };

	out[0] = colour->red;
	out[1] = colour->green;
	out[2] = colour->blue;
}

static int render_frame(LBX_IMG *img, GdkPixbuf *buf, unsigned frame)
{
	unsigned char **imgdata, **imgmask;
	struct lbx_imginfo info;
	unsigned char *data;
	unsigned stride;

	lbximg_getinfo(image, &info);
	imgdata = lbximg_getframe(image, frame);
	if (!imgdata) {
		printf("failed to decode frame %u\n", frame);
		return -1;
	}
	imgmask = lbximg_getmask(image);

	data   = gdk_pixbuf_get_pixels(buf);
	stride = gdk_pixbuf_get_rowstride(buf);

	for (unsigned i = 0; i < info.height; i++) {
		unsigned char (*p)[3] = (void *)(data + i*stride);

		for (unsigned j = 0; j < info.width; j++) {
			if (imgmask[i][j]) {
				get_colour(imgdata[i][j], p[j]);
			} else {
				p[j][0] = 0;
				p[j][1] = 0;
				p[j][2] = 0;
			}
		}
	}

	return 0;
}

void set_frame(GtkSpinButton *spin, gpointer data)
{
	unsigned frame = gtk_spin_button_get_value_as_int(spin);

	render_frame(image, framebuf, frame);
	gdk_window_invalidate_rect(canvas->window, NULL, FALSE);
}

static void tick(void *p, double delta)
{
	static double elapsed = 0;
	double seconds_per_frame = 1.0/15;
	struct lbx_imginfo info;
	GtkSpinButton   *spin;
	GtkToggleButton *play;
	unsigned frame;

	if (!image)
		return;

	lbximg_getinfo(image, &info);
	spin = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "framespin"));
	play = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "playbutton"));

	frame = gtk_spin_button_get_value_as_int(spin);
	if (!gtk_toggle_button_get_active(play))
		return;

	elapsed += delta;
	while (elapsed > seconds_per_frame) {
		elapsed -= seconds_per_frame;

		if (++frame >= info.nframes) {
			if (!info.looping) {
				gtk_toggle_button_set_active(play, FALSE);
				break;
			}

			frame = info.loopstart;
		}

		gtk_spin_button_set_value(spin, frame);
	}
}

static gboolean timeout(gpointer data)
{
	static double lasttime = INFINITY;
	static GTimer *timer;

	if (isinf(lasttime)) {
		timer = g_timer_new();
		if (timer) {
			lasttime = 0;
		}
	} else {
		double time = g_timer_elapsed(timer, NULL);

		tick(data, time - lasttime);
		lasttime = time;
	}

	return TRUE;
}

gboolean canvas_expose(GtkWidget *canvas, GdkEventExpose *event, gpointer data)
{
	if (!framebuf)
		return TRUE;

	gdk_draw_pixbuf(canvas->window, NULL, framebuf,
		event->area.x, event->area.y, event->area.x, event->area.y,
		event->area.width, event->area.height,
		GDK_RGB_DITHER_NORMAL, 0, 0);

	return TRUE;
}

static int init_framedata(void)
{
	struct lbx_imginfo info;
	GtkSpinButton *spin;

	lbximg_getinfo(image, &info);
	framebuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8,
		info.width, info.height);
	if (!framebuf) {
		printf("failed to allocate pixbuf\n");
		return -1;
	}

	gtk_widget_set_size_request(canvas, info.width, info.height);

	spin = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "framespin"));
	gtk_spin_button_set_range(spin, 0, info.nframes-1);
	gtk_spin_button_set_value(spin, 0);
	set_frame(spin, NULL);
	return 0;
}

static int img_close(void *handle)
{
	lbx_file_close(handle);
	return 0;
}

void set_image_data(GtkComboBox *combo)
{
	GtkTreeIter iter;
	gpointer lbx;
	guint    index;
	LBXfile  *f;

	if (!gtk_combo_box_get_active_iter(combo, &iter))
		return;

	gtk_tree_model_get(GTK_TREE_MODEL(archives), &iter,
		1, &lbx,
		2, &index,
		-1);

	f = lbx_file_open(lbx, index);
	if (f) {
		struct lbx_imginfo info;

		image = lbximg_open(f, &lbx_arch_fops, img_close);
		if (image) {
			memset(palette_internal, 0, sizeof palette_internal);
			if (lbximg_getpalette(image, palette_internal) == -1) {
				puts("crap");
				lbximg_close(image);
			}

			if (init_framedata() == -1) {
				puts("crap");
				lbximg_close(image);
			}
		} else {
			puts("fuck");
			lbx_file_close(f);
		}
	} else {
		puts("damn");
	}
}

void show_about(GtkWidget *widget)
{
	GtkWindow *parent = GTK_WINDOW(gtk_widget_get_toplevel(widget));

	gtk_show_about_dialog(parent,
		"name", g_get_application_name(),
		"version", PACKAGE_VERSION,
		"copyright", "Copyright \xc2\xa9 2010 Nick Bowler",
		"website", "http://toom.sourceforge.net/",
		NULL);
}

int load_archive(const char *path)
{
	gchar *basename;
	GtkTreeIter iter1, iter2;
	unsigned nfiles;
	LBX *lbx;

	lbx = lbx_fopen(path);
	if (!lbx)
		return -1;

	basename = g_path_get_basename(path);
	gtk_tree_store_append(archives, &iter1, NULL);
	gtk_tree_store_set(archives, &iter1,
		0, basename,
		1, (gpointer)lbx,
		-1);
	g_free(basename);

	nfiles = lbx_numfiles(lbx);
	for (unsigned i = 0; i < nfiles; i++) {
		struct lbx_statbuf stat;

		lbx_stat(lbx, i, &stat);
		gtk_tree_store_append(archives, &iter2, &iter1);
		gtk_tree_store_set(archives, &iter2,
			0, stat.name,
			1, (gpointer)lbx,
			2, (guint)i,
			-1);
	}

	return 0;
}

void load_error(const char *name, GtkWindow *window)
{
	GtkWidget *dialog;

	dialog = gtk_message_dialog_new(window,
		GTK_DIALOG_MODAL,
		GTK_MESSAGE_ERROR,
		GTK_BUTTONS_OK,
		"Failed to open %s.  Decent error messages coming eventually.",
		name);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

void open_archive(GtkWidget *widget)
{
	GtkWindow *window = GTK_WINDOW(gtk_widget_get_toplevel(widget));
	GtkFileFilter *lbxfilter, *allfilter;
	GtkFileChooser *chooser;
	GtkTreeIter iter;

	chooser = GTK_FILE_CHOOSER(gtk_file_chooser_dialog_new(
		"Load Archive",
		window,
		GTK_FILE_CHOOSER_ACTION_OPEN,
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
		NULL));

	lbxfilter = gtk_file_filter_new();
	gtk_file_filter_set_name(lbxfilter, "LBX Archives (*.lbx)");
	gtk_file_filter_add_pattern(lbxfilter, "*.lbx");

	allfilter = gtk_file_filter_new();
	gtk_file_filter_set_name(allfilter, "All Files");
	gtk_file_filter_add_pattern(allfilter, "*");

	gtk_file_chooser_add_filter(chooser, lbxfilter);
	gtk_file_chooser_add_filter(chooser, allfilter);

	while (gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_ACCEPT) {
		gchar *name = gtk_file_chooser_get_filename(chooser);

		if (load_archive(name) == 0) {
			g_free(name);
			break;
		}

		load_error(name, GTK_WINDOW(chooser));
		g_free(name);
	}

	gtk_widget_destroy(GTK_WIDGET(chooser));
}

static void interface_init(void)
{
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GtkComboBox *combo;
	GtkTreeView *tree;

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_add_attribute(column, renderer, "text", 0);

	tree = GTK_TREE_VIEW(gtk_builder_get_object(builder, "treeview"));
	gtk_tree_view_set_model(tree, GTK_TREE_MODEL(archives));
	gtk_tree_view_append_column(tree, column);

	combo = GTK_COMBO_BOX(gtk_builder_get_object(builder, "imagechooser"));
	gtk_combo_box_set_model(combo, GTK_TREE_MODEL(archives));
	gtk_combo_box_entry_set_text_column(GTK_COMBO_BOX_ENTRY(combo), 0);
}

int main(int argc, char **argv)
{
	GtkWidget *window;
	GError *err = NULL;

	if (!gtk_init_with_args(&argc, &argv, NULL, NULL, NULL, &err)) {
		fprintf(stderr, "%s\n", err->message);
		return EXIT_FAILURE;
	}

	archives = gtk_tree_store_new(3,
		G_TYPE_STRING, G_TYPE_POINTER, G_TYPE_UINT);
	while (*argv)
		load_archive(*argv++);

	builder = gtk_builder_new();
	if (!gtk_builder_add_from_file(builder, "src/gui/lbxgui.glade", &err)) {
		fprintf(stderr, "%s\n", err->message);
		return EXIT_FAILURE;
	}

	window = GTK_WIDGET(gtk_builder_get_object(builder, "mainwindow"));
	canvas = GTK_WIDGET(gtk_builder_get_object(builder, "canvas"));

	interface_init();
	gtk_builder_connect_signals(builder, window);

	g_timeout_add(33, timeout, NULL);
	gtk_widget_show_all(window);
	gtk_main();
}
